      program square 

c
c     program author: Alan Shapiro
c
c     This program computes the analytical solution for streamfunction, 
c     buoyancy, vertical velocity, and vorticity for steady two-dimensional
c     flow over an inhomogeneously heated lower surface. The surface
c     buoyancy is in the form of a square wave (step function). The 
c     dependent variables satisfy the steady linearized viscous Boussinesq
c     equations of motion, thermal energy, and mass conservation 
c     (impermeability condition). The parameters used in this particular
c     test correspond to test "A-1" as described in the manuscript
c     "An analytical verification test for numerically simulated convective
c     flow above a thermally heterogeneous surface" in review at GMD. 
c
c     All quantities are dimensional.
c
c     To compile and run the code on Sardine (Intel Xeon Mac Pro):
c
c                                  ifort -o x.square square.f
c                                  ./x.square > square.out &
c
c     The following parameters must be specified (they are declared
c     and specified after the implicit none statement). The values 
c     listed here correspond to test A1:
c
c     nz = 1025        # of grid points in z direction. So 1024 intervals.
c     dz = 0.01        vertical grid spacing in meters
c     nx = 513         # of grid points in x direction. So 512 intervals.
c     dx = 0.01        horizontal grid spacing in meters
c     nu = 0.001       kinematic viscosity
c     alp = 0.001      thermal diffusivity, alpha
c     BV = 0.02        Brunt-Vaisala frequency N
c     bmax = 0.00001   max surface buoyancy  
c     bmin = -0.00001  min surface buoyancy; be sure to set bmin = -bmax 
c     nmax = 50000     number of terms in the Fourier series
c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

      implicit none

c
c     Declare variables and parameters.
c
 
      integer j                 ! Grid point index for z coordinate
      integer nz                ! Number of grid points in z direction
      parameter (nz = 1025)    
      real z(nz)                ! z coordinate
      real dz                   ! Grid spacing in z direction
      parameter (dz = 0.01)
      real (kind=8) dz2         ! dz**2

      integer i                 ! Grid point index for x coordinate
      integer nx                ! Number of grid points in x direction
      parameter (nx = 513)    
      real x(nx)                ! x coordinate
      real dx                   ! Grid spacing in x direction
      parameter (dx = 0.01)
      real (kind=8) dx2         ! dx**2

      real nu                   ! Kinematic viscosity 
      parameter (nu = 0.001)
  
      real alp                  ! Thermal diffusivity
      parameter (alp = 0.001)

      real BV                   ! Brunt-Vaisala frequency 
      parameter (BV = 0.02)
      real (kind=8) BV43      

      real bmax                 ! Maximum buoyancy
      parameter (bmax = 0.00001)   
    
      real bmin                 ! Minimum buoyancy
      parameter (bmin = -0.00001)   

      real L                    ! Wavelength of square wave

      integer n                 ! Wavenumber n in Fourier series
      integer nmax              ! Number of terms in the Fourier series
      parameter (nmax = 50000) 

      real (kind=8) bn          ! n'th Fourier coefficient

      real (kind=8) k           ! Wavenumber k is related to n by: k = n*pi/L
 
      real (kind=8) M0          ! First vertical wavenumber, M0 
      real (kind=8) r           ! Appears in solution for M1 and M2
      real (kind=8) phi         ! Appears in solution for M1 and M2
      real (kind=8) mu          ! mu = M0/sqrt(r)
 
      real (kind=8) Zc          ! Zc = z*sqrt(r)*cos(phi/2.)
      real (kind=8) Zs          ! Zs = z*sqrt(r)*sin(phi/2.)

      real b(nx, nz)            ! Buoyancy
      real psi(nx, nz)          ! Streamfunction 
      real w(nx, nz)            ! Vertical velocity 
      real eta(nx, nz)          ! Vorticity 
      real etaFD(nx, nz)        ! Vorticity from finite dif of Lap(psi)
      real u(nx, nz)            ! Horizontal velocity 
      real uFD(nx, nz)          ! Horizontal velocity from finite dif

      real (kind=8) Lapb(nx, nz)
      real (kind=8) Lapeta(nx, nz)
      real (kind=8) dbdx(nx, nz)
      real (kind=8) arg 
      real (kind=8) arg1
      real (kind=8) arg2
      real (kind=8) exp1
      real (kind=8) exp2
      real (kind=8) pi
      real (kind=8) cs
      real (kind=8) sn
      real (kind=8) numer
      real (kind=8) denom 
      real (kind=8) fac
      real nonlin
      real nonlinmax
      real dbdxmax
      real Lapbmax
      real rat
      
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

      print *, ' '
      print *, '***** Welcome to program "square" ***** '
      print *, ' '

      L = (nx - 1.)*dx

      print *, 'nu = ', nu 
      print *, 'alp = ', alp 
      print *, 'BV = ', BV 
      print *, 'bmax = ', bmax
      print *, 'bmin = ', bmin
      print *, 'wavelength L = ', L
      print *, '# terms in Fourier series, nmax =', nmax
      print *, 'nz = ', nz 
      print *, 'nx = ', nx
      print *, 'dz = ', dz 
      print *, 'dx = ', dx

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
 
c
c     Set up the computational domain. 
c 

      do 100 j = 1, nz
        z(j) = (j - 1.0)*dz
100   continue

c     print *, ' '
c     print *, 'i, x '
      do 200 i = 1, nx
        x(i) = (i - 1.0)*dx
200   continue

      dx2 = dx*dx
      dz2 = dz*dz

      pi = 4.*atan(1.)

c
c     Calculate b, psi, w, and vorticity (eta).
c

      do 500 i = 1, nx 
      do 500 j = 1, nz 

        b(i, j) = 0.           ! initialize b
        psi(i, j) = 0.         ! initialize psi
        w(i, j) = 0.           ! initialize w
        eta(i, j) = 0.         ! initialize eta

        do 300 n = 1, nmax     ! Start of Fourier loop

        k = n*pi/L                ! Define k

        bn = bmax*(1.-cos(n*pi/2.)) + bmin*(cos(n*pi/2.)-cos(n*pi))
        bn = 2.*bn/(n*pi)

c
c       Compute M0, r, and phi   
c
   
        arg = 2.*pi/3.
        cs = cos(arg) 
        sn = sin(arg) 
        numer = (BV*k)**(2./3.)
        denom = (nu*alp)**(1./3.)

        M0 = - sqrt(k**2 + numer/denom)

        r = sqrt( (k**2 + numer*cs/denom)**2 + (numer*sn/denom)**2)

        phi = acos( (k**2 + numer*cs/denom)/r )

        mu = M0/sqrt(r)

        BV43 = BV**(4./3.)
        denom = mu + 2.*cos(pi/3. + phi/2.)

        Zs = z(j)*sqrt(r)*sin(phi/2.)
        Zc = z(j)*sqrt(r)*cos(phi/2.)

        arg1 = - Zc
        if (arg1 .lt. -200.) then
          exp1 = 0.
        else
          exp1 = exp(arg1)
        end if

     
        arg2 = M0*z(j) 
        if (arg2 .lt. -200.) then
          exp2 = 0.
        else
          exp2 = exp(arg2)
        end if

        numer = exp1*(mu*cos(Zs+pi/6.) + cos(Zs+pi/6.+phi/2.))
        numer = numer - exp2*sin(phi/2.)
        fac = 2.*bn/sqrt(3.)
        b(i, j) = b(i, j) + fac * numer * sin(k*x(i))/denom
     
        numer = exp1*(mu*sin(Zs) + sin(Zs+phi/2.))
        numer = numer - exp2*sin(phi/2.)
        fac = (2.*bn*alp**(2./3.))/(sqrt(3.)*BV43*(k*nu)**(1./3.))
        psi(i, j) =  psi(i, j) + fac * numer * cos(k*x(i))/denom

        fac = (2.*bn*(alp*k)**(2./3.))/(sqrt(3.)*BV43*nu**(1./3.))
        w(i, j) =  w(i, j) + fac * numer * sin(k*x(i))/denom
   
        numer = exp1*(mu*cos(Zs-pi/6.) + cos(Zs-pi/6.+phi/2.))
        numer = numer + exp2*sin(phi/2.)
        fac = -(2.*bn*(alp*k)**(1./3.))/(sqrt(3.)*(BV*nu)**(2./3.))
        eta(i, j) = eta(i, j) + fac * numer * cos(k*x(i))/denom

        numer = exp1*(mu*sin(phi/2. - Zs) - sin(Zs))
        numer = numer - mu*exp2*sin(phi/2.)
        fac = (2.*bn*sqrt(r/3.)*alp**(2./3.))/(BV43*(k*nu)**(1./3.))
        u(i, j) =  u(i, j) + fac * numer * cos(k*x(i))/denom

300   continue      ! End of Fourier loop

500   continue

c
c     Finite difference psi to get uFD. Skip top and bottom points.
c

      do 600 i = 1, nx 
      do 600 j = 2, nz-1 
        uFD(i, j) = (psi(i, j+1) - psi(i, j-1))/(2.*dz)
600   continue

c
c     Verify that surface buoyancy computed from the Fourier series
c     is really a square wave.
c

      print *, ' '
      print *, 'Verify that surface buoyancy is really a square wave'
      print *, ' '

      do 900 i = 1, nx 
        print *, 'i, x, bs  = ', i, x(i), b(i,1)
900   continue

c
c     Verify that the finite difference form of vorticity (etaFD 
c     obtained as the finite difference form of Laplacian of analytical
c     streamfunction) agrees with purely analytical form of vorticity.
c     Also compare finite difference u (from psi) and analytical u
c

      do 1000 i = 2, nx-1 
      do 1000 j = 2, nz-1
        etaFD(i, j) =  (psi(i+1,j) - 2.*psi(i,j) + psi(i-1,j))/dx2 
     1               + (psi(i,j+1) - 2.*psi(i,j) + psi(i,j-1))/dz2 
1000  continue

      print *, ' '
      print *, 'Compare finite dif eta, u with analytical eta, u'
      print *, ' '

      do 1100 j = 2, nz-1
        print *, ' j = ', j
        print *, 'etaFD(3,j), eta(3,j) =', etaFD(3,j), eta(3,j)
        print *, 'uFD(3,j), u(3,j) =', uFD(3,j), u(3,j)
        print *, ' '
1100  continue

c
c     Verify that solution is in a linear regime, i.e., that the nonlinear
c     terms in the vorticity and thermal energy equations are much smaller
c     than the corresponding linear terms.
c
c
c     Check that the vorticity advection, u*d(eta)/dx + w*d(eta)/dz,
c     is much smaller than the baroclinic generation, db/dx.
c

      do 1500 i = 2, nx-1 
      do 1500 j = 2, nz-1
        Lapb(i,j) =  (b(i+1,j) - 2.*b(i,j) + b(i-1,j))/dx2 
     1             + (b(i,j+1) - 2.*b(i,j) + b(i,j-1))/dz2 
        dbdx(i,j) =  (b(i+1, j) - b(i-1, j))/(2.*dx)
1500  continue

      print *, ' '
      print *, 'Linear regime check of vorticity equation'
      print *, ' '

      dbdxmax = 0.
      nonlinmax = 0.

      do 1700 i = 2, nx-1     
      do 1700 j = 2, nz-1

        nonlin =  u(i,j)*(eta(i+1,j) - eta(i-1,j))/(2.*dx) +
     1            w(i,j)*(eta(i,j+1) - eta(i,j-1))/(2.*dz)

        nonlin = abs(nonlin) 

        if (abs(dbdx(i,j)) .gt. dbdxmax) then
          dbdxmax = abs(dbdx(i,j))
        end if

        if (nonlin .gt. nonlinmax) then
          nonlinmax = nonlin
        end if  

1700  continue

      rat = nonlinmax/dbdxmax

      print *, 'nonlinmax =', nonlinmax, 'dbdxmax =', dbdxmax
      print *, 'ratio =', rat

c
c     Now check that the thermal advection, u*db/dx + w*db/dz, is much
c     smaller than the thermal diffusion, alp * Lapb. 
c

      print *, ' '
      print *, 'Linear regime check of thermal energy equation'
      print *, ' '

      Lapbmax = 0.
      nonlinmax = 0.

      do 1800 i = 2, nx-1     
      do 1800 j = 2, nz-1

        nonlin =  u(i,j)*(b(i+1,j) - b(i-1,j))/(2.*dx) + 
     1            w(i,j)*(b(i,j+1) - b(i,j-1))/(2.*dz)
        nonlin = abs(nonlin) 

        if (abs(Lapb(i,j)) .gt. Lapbmax) then
          Lapbmax = abs(Lapb(i,j))
        end if

        if (nonlin .gt. nonlinmax) then
          nonlinmax = nonlin
        end if  

1800  continue

      Lapbmax = alp*Lapbmax

      rat = nonlinmax/Lapbmax

      print *, 'nonlinmax =', nonlinmax, 'alp*Lapbmax =', Lapbmax
      print *, 'ratio =', rat

c
c     Output solution for later use in post-processing/graphics codes. 
c

      open(140, file = 'w.dat', status = 'unknown')
      open(145, file = 'u.dat', status = 'unknown')
      open(150, file = 'b.dat', status = 'unknown')
      open(160, file = 'psi.dat', status = 'unknown')
      open(170, file = 'eta.dat', status = 'unknown')

      do 2000 i = 1, nx 
      do 2000 j = 1, nz 
        write(140, 70) x(i), z(j), w(i,j)
        write(145, 70) x(i), z(j), u(i,j)
        write(150, 70) x(i), z(j), b(i,j)
        write(160, 70) x(i), z(j), psi(i,j)
        write(170, 70) x(i), z(j), eta(i,j)
2000  continue

70    format(3(1x, e14.7))

      close(140)
      close(145)
      close(150)
      close(160)
      close(170)

      stop
      end
